<html>
<!--{{{SGML}}}--><!--{{{--><head>
<title>A tool to find your way in C sources</title>
<!--}}}--></head>
<!--{{{--><body>
<h1>A tool to find your way in C sources</h1>
<hr>
<a href="http://www.unipress.com/att/new/cscope.html"><code>cscope</code></a>
is a commercial tool, available from AT&T.  It creates an index file
from C sources and allows interactive queries like listing all lines
which call a specific function or identifier.
<p>
Although it is a very handy tool, there is one problem: Student purses
cringe away from the word `commercial' alone.  :-/ As a conclusion, I
decided to implement a free and small clone which implements certain
functions of cscope: cs.
<p>
cs offers the following functions:
<ul>
<li>find C symbol
<li>find functions called from a certain function
<li>find functions calling a certain function
<li>find string literals
<li>find files #including a certain header file
<li>find file
<li>find POSIX naming violations
<li>execute shell command
</ul>
The found lines will be displayed.  Besides browsing through them, cs
allows to edit them using your favourite editor.  If you frequently
grep for identifiers and edit the found files, then cs is for you.
cs should work on any system which supports the following:
<ul>
<li>POSIX
<li>ANSI C
<li>SYSV curses library
</ul>
<!--}}}--></body>
</html>
